# -*- coding: utf-8 -*-

"""Storage for detected and fetched widgets."""

from __future__ import annotations

__all__ = ['PartnerPool', 'POOL', 'UPDATER']

import json
import os
import typing
from urllib import parse
import yaml
from qtpy import QtCore
from anaconda_navigator import config
from anaconda_navigator.utils import extra_collections
from anaconda_navigator.utils.logs.loggers import logger
from anaconda_navigator.utils import workers
from . import core
from . import resources
from . import simple_advertisement

if typing.TYPE_CHECKING:
    import typing_extensions
    from qtpy import QtWidgets

    class PartnerConfiguration(typing_extensions.TypedDict, total=False):
        """Dictionary of an additional configuration, that might be used by advertisements and the application."""

        utm_campaign: str


def __init_configuration_paths() -> typing.MutableSequence[str]:
    """Initialize sequence of paths to search configurations in."""
    result: typing.List[str] = []

    if config.WIN:
        # pylint: disable=import-outside-toplevel
        from anaconda_navigator.external.knownfolders import get_folder_path, FOLDERID  # type: ignore
        result.extend((
            os.path.join(get_folder_path(FOLDERID.ProgramData)[0], 'Anaconda3', 'etc', 'partner.yml'),
            os.path.join(get_folder_path(FOLDERID.ProgramData)[0], 'Miniconda3', 'etc', 'partner.yml'),
            os.path.join(os.path.expanduser('~'), 'Anaconda3', 'etc', 'partner.yml'),
            os.path.join(os.path.expanduser('~'), 'Miniconda3', 'etc', 'partner.yml'),
        ))

    else:
        result.extend((
            os.path.join('/', 'etc', 'anaconda', 'partner.yml'),
            os.path.join('opt', 'anaconda3', 'etc', 'partner.yml'),
            os.path.join('opt', 'miniconda3', 'etc', 'partner.yml'),
            os.path.join(os.path.expanduser('~'), 'anaconda3', 'etc', 'partner.yml'),
            os.path.join(os.path.expanduser('~'), 'miniconda3', 'etc', 'partner.yml'),
        ))

    result.extend((
        os.path.join(config.CONF_PATH, 'partner.yml'),
    ))

    return result


CONFIGURATION_PATHS: 'typing_extensions.Final[typing.Sequence[str]]' = __init_configuration_paths()


class PartnerPool(typing.Mapping[core.PartnerWidgetPlacement, 'QtWidgets.QWidget']):
    """Collection of widgets for partner attribution."""

    __slots__ = ('__content', '__settings')

    def __init__(self) -> None:
        """Initialize new :class:`~PartnerPool` instance."""
        self.__settings: core.PartnerSettings = core.PartnerSettings()
        self.__content: 'typing_extensions.Final[typing.Dict[core.PartnerWidgetPlacement, core.PartnerWidget]]' = {}

    @property
    def settings(self) -> core.PartnerSettings:  # noqa: D401
        """Common configuration that might be used by widgets and different application components."""
        return self.__settings

    def _set_settings(self, value: typing.Union[core.PartnerSettings, 'core.SettingsDict']) -> None:
        """Update value of the `configuration`."""
        if not isinstance(value, core.PartnerSettings):
            value = core.PartnerSettings(value)
        self.__settings = value

    def register(self, widget: core.PartnerWidget) -> None:
        """Register new widget in the pool."""
        self.__content[widget.placement] = widget

    def __getitem__(self, key: core.PartnerWidgetPlacement) -> typing.Optional[QtWidgets.QWidget]:
        """Retrieve widget that should be placed in `key` location."""
        value: typing.Optional[core.PartnerWidget] = self.__content.get(key, None)
        if value is None:
            return None
        return value.widget()

    def __iter__(self) -> typing.Iterator[core.PartnerWidgetPlacement]:
        """Iterate through available location options."""
        return iter(core.PartnerWidgetPlacement)

    def __len__(self) -> int:
        """Total number of """
        return len(core.PartnerWidgetPlacement)


POOL: 'typing_extensions.Final[PartnerPool]' = PartnerPool()


class Updater(QtCore.QObject):
    """Contain update utm parameters functional"""

    sig_updated = QtCore.Signal()

    @staticmethod
    def __read_partner_configuration(paths: typing.Optional[typing.Sequence[str]] = None) -> 'PartnerConfiguration':
        """Check exists file of partner config or return empty default"""
        if paths is None:
            paths = CONFIGURATION_PATHS

        path: str
        for path in paths:
            try:
                stream: typing.TextIO
                with open(path, 'rt', encoding='utf8') as stream:
                    result: 'PartnerConfiguration' = yaml.safe_load(stream)

                    if not isinstance(result, typing.Mapping):
                        logger.exception('broken partner configuration file at %s', path)  # type: ignore
                        continue

                    return result

            except FileNotFoundError:
                continue

            except (OSError, yaml.YAMLError):
                logger.exception('broken partner configuration file at %s', path)
                continue

        return {}

    @workers.Task
    def update(self) -> None:
        """Fetch and apply advertisement configuration to the :data:`~POOL`."""
        configuration: core.ConfigurationDict = {
            'widgets': [],
            'settings': {
                'url_parameters': {},
            },
        }

        # prepare a list of URLs to fetch advertisement information from
        advertisement_urls: extra_collections.OrderedSet[str] = extra_collections.OrderedSet()

        partner_configuration: 'PartnerConfiguration' = self.__read_partner_configuration()
        try:
            advertisement_urls.add(
                f'https://www.anaconda.com/api/navigator/partner/{parse.quote(partner_configuration["utm_campaign"])}',
            )
        except KeyError:
            pass

        advertisement_urls.add(config.CONF.get('main', 'advertisement_url'))
        advertisement_urls.add('https://www.anaconda.com/api/navigator')

        # process advertisement URLs
        advertisement_url: str
        for advertisement_url in advertisement_urls:
            try:
                advertisement_content: typing.Optional[bytes] = resources.load_resource(advertisement_url)
                if not advertisement_content:
                    continue

                # Fetch configuration
                definition: 'simple_advertisement.CompositeAdvertisementDict' = {
                    'placement': core.PartnerWidgetPlacement.BOTTOM_LEFT_CORNER,
                    'advertisements': json.loads(advertisement_content),
                }
                configuration['widgets'].append(definition)

                # Extract utm parameters from the redirect_url
                item: simple_advertisement.SimpleAdvertisementDict
                for item in definition['advertisements']:
                    if 'redirect_url' not in item:
                        continue

                    configuration['settings']['url_parameters'].update(
                        (key, value)
                        for key, value in parse.parse_qsl(parse.urlparse(item['redirect_url']).query)
                        if key.startswith('utm_')
                    )
                    break  # NOTE: remove if we need to combine data from all URLs and not just the first valid one

                # Remember the working API url
                config.CONF.set('main', 'advertisement_url', advertisement_url)
                break

            except BaseException:  # pylint: disable=broad-except
                logger.exception('failed to initialize advertisement')

        # apply `utm_campaign` from `partner_configuration`
        try:
            configuration['settings']['url_parameters']['utm_campaign'] = partner_configuration['utm_campaign']
        except KeyError:
            pass

        # apply default values for required parameters
        configuration['settings']['url_parameters'].setdefault('utm_source', 'anaconda_navigator')

        # make configuration available to everything
        POOL._set_settings(configuration['settings'])  # pylint: disable=protected-access

        # create widgets according to the configuration
        for widget in configuration['widgets']:
            instance: core.PartnerWidget

            # NOTE: to support additional built-in advertisement types:
            #
            #       if widget.get('type', 'simple') == 'simple':
            instance = simple_advertisement.CompositeAdvertisementWidget(
                definition=typing.cast('simple_advertisement.CompositeAdvertisementDict', widget),
                settings=POOL.settings,
            )

            POOL.register(instance)

        # NOTE: NAV-753 should be here
        #
        #       Parse other locations for widgets, initialize them with `POOL.settings` and then `POOL.register` them.

        self.sig_updated.emit()


UPDATER: 'typing_extensions.Final[Updater]' = Updater()



# -*- coding: utf-8 -*-

"""Core components for storing and executing issue solvers."""

from __future__ import annotations

__all__ = [
    'SolvedError', 'SolverCollection',
    'ConfigurationContext', 'CONFIGURATION_POOL',
    'ConflictContext', 'CONFLICT_POOL',
]

import typing
from anaconda_navigator.utils import solvers as common_solvers

if typing.TYPE_CHECKING:
    import typing_extensions
    from anaconda_navigator.api import anaconda_api
    from anaconda_navigator.config import user as user_config

    TContext_contra = typing.TypeVar('TContext_contra', contravariant=True)

    class Solver(typing_extensions.Protocol[TContext_contra]):  # pylint: disable=too-few-public-methods
        """
        Common interface for issue solvers.

        Each solver should:

        - check if there is an issue
        - perform corresponding actions to fix the issue
        - report what was wrong and what was changed
        """

        def __call__(self, context: TContext_contra) -> typing.Union[None, SolvedError, typing.Iterable[SolvedError]]:
            """
            Solve an issue if it is detected.

            :param context: Data structure with data that should be checked for issue and updated if one is found.
            :return: Message(s) about what was fixed.
            """


TInvContext = typing.TypeVar('TInvContext')


class SolvedError(typing.NamedTuple):
    """Details on solved issue."""

    caption: str
    message: str
    tags: typing.Sequence[str]


class SolverCollection(common_solvers.SolverCollection['Solver[TInvContext]'], typing.Generic[TInvContext]):
    """Collection of issue solvers."""

    __slots__ = ()

    def solve(
            self,
            context: TInvContext,
            *,
            tags: typing.Union[None, str, typing.Iterable[str]] = None,
    ) -> typing.Iterator[SolvedError]:
        """
        Detect and solve issues for the `context`.

        :param context: Context to check for issues and fix them.
        :param tags: Limit to issue solvers with specific tags.
        :return: Iterator of details about solved issues.
        """
        solver_record: common_solvers.SolverRecord['Solver[TInvContext]']
        for solver_record in self.only(tags=tags):
            result: typing.Union[None, SolvedError, typing.Iterable[SolvedError]] = solver_record.solver(context)
            if result is None:
                continue
            if isinstance(result, SolvedError):
                yield result
            else:
                yield from result


class ConfigurationContext(typing.NamedTuple):
    """Context for fixing a general configuration issues."""

    api: 'anaconda_api._AnacondaAPI'
    config: 'user_config.UserConfig'


CONFIGURATION_POOL: 'typing_extensions.Final[SolverCollection[ConfigurationContext]]' = SolverCollection()


class ConflictContext(typing.NamedTuple):
    """Context for fixing issues with conflicting Navigator and Conda configurations."""

    api: 'anaconda_api._AnacondaAPI'
    config: 'user_config.UserConfig'
    conda_info: typing.Any


CONFLICT_POOL: 'typing_extensions.Final[SolverCollection[ConflictContext]]' = SolverCollection()



# -*- coding: utf-8 -*-

"""Additional utility functions to use with pycharm."""

from __future__ import annotations

__all__ = [
    'VSCodeVersionChecker', 'vscode_detector', 'vscode_extra_arguments', 'vscode_install_extensions',
    'vscode_update_config',
]

import datetime
import json
import os
import typing
from anaconda_navigator import config as navigator_config
from anaconda_navigator.api import conda_api
from anaconda_navigator.utils.conda import launch as conda_launch_utils
from anaconda_navigator.utils.logs import logger
from .. import detectors
from .. import validation_utils
from . import detector_utils

if typing.TYPE_CHECKING:
    from anaconda_navigator.api import process
    from .. import base
    from .. import parsing_utils


class VSCodeVersionChecker(detectors.Filter):  # pylint: disable=too-few-public-methods
    """Detect version of the VS Code application."""

    __slots__ = ()

    @classmethod
    def _parse_configuration(cls, *args: typing.Any, **kwargs: typing.Any) -> detectors.Detector:
        """Parse configuration for this particular :class:`~Detector`."""
        validation_utils.has_items(at_most=0)(args, field_name='args')

        validation_utils.mapping_is_empty()(kwargs)

        with validation_utils.catch_exception():
            return cls()

    def __call__(
            self,
            parent: typing.Iterator[detectors.DetectedApplication],
            *,
            context: detectors.DetectorContext,
    ) -> typing.Iterator[detectors.DetectedApplication]:
        """Iterate through detected applications."""
        application: 'detectors.DetectedApplication'
        for application in parent:
            if not application.executable:
                continue

            stdout: str
            stdout, _, _ = conda_launch_utils.run_process([application.executable, '--version'])
            if stdout:
                yield application.replace(version=stdout.splitlines()[0])
            else:
                yield application.replace(version='unknown')


def vscode_detector() -> detectors.Source:
    """Prepare detector for VS Code."""
    return detectors.Group(
        detectors.OsXOnly(
            detectors.Group(
                detectors.CheckConfiguredRoots(),
                detectors.CheckKnownOsXRoots('Visual Studio Code.app'),
                detectors.AppendExecutable(
                    detectors.join('Contents', 'Resources', 'app', 'bin', 'code'),
                ),
            ),

            detectors.Group(
                detectors.CheckPATH('code'),
                detectors.AppendRoot(level=4),
            ),
        ),

        detectors.LinuxOnly(
            detectors.Group(
                detectors.CheckConfiguredRoots(),
                detectors.CheckKnownRoots(
                    detectors.join(
                        detectors.FOLDERS['linux/root'], 'usr', 'share', 'code',
                    ),
                    detectors.join(
                        detectors.FOLDERS['linux/snap_primary'], 'code', 'current', 'usr', 'share', 'code',
                    ),
                    detectors.join(
                        detectors.FOLDERS['linux/snap_secondary'], 'code', 'current', 'usr', 'share', 'code',
                    ),
                ),
                detectors.AppendExecutable(
                    detectors.join('bin', 'code'),
                ),
            ),

            detectors.Group(
                detectors.CheckPATH('code'),
                detectors.AppendRoot(level=1),
            ),
        ),

        detectors.WindowsOnly(
            detectors.Group(
                detectors.CheckConfiguredRoots(),
                detectors.CheckKnownRoots(
                    detectors.join(detectors.FOLDERS['windows/program_files_x86'], 'Microsoft VS Code'),
                    detectors.join(detectors.FOLDERS['windows/program_files_x64'], 'Microsoft VS Code'),
                    detectors.join(detectors.FOLDERS['windows/local_app_data'], 'Programs', 'Microsoft VS Code'),
                ),
                detectors.AppendExecutable(
                    detectors.join('bin', 'code.cmd'),
                ),
            ),

            detectors.Group(
                detectors.CheckPATH('code.cmd'),
                detectors.AppendRoot(level=1),
            ),

            detectors.Group(
                detectors.CheckRootInWindowsRegistry(
                    detectors.RegistryKey(
                        root='HKEY_CURRENT_USER',
                        key=(
                            'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\'
                            '{771FD6B0-FA20-440A-A002-3B3BAC16DC50}_is1'
                        ),
                        sub_key='Inno Setup: App Path',
                    ),
                    detectors.RegistryKey(
                        root='HKEY_LOCAL_MACHINE',
                        key=(
                            'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\'
                            '{EA457B21-F73E-494C-ACAB-524FDE069978}_is1'
                        ),
                        sub_key='Inno Setup: App Path',
                    ),
                ),
                detectors.AppendExecutable(
                    detectors.join('bin', 'code.cmd'),
                ),
            ),

            detectors.Group(
                detectors.CheckExecutableInWindowsRegistry(
                    detectors.RegistryKey(
                        root='HKEY_CLASSES_ROOT',
                        key='vscode\\shell\\open\\command',
                        converter=detector_utils.extract_app_from_command,
                    ),
                ),
                detectors.AppendRoot(level=1),
            ),
        ),

        VSCodeVersionChecker(),
    )


def vscode_extra_arguments() -> typing.Sequence[str]:
    """Return default extra arguments for vscode."""
    return '--user-data-dir', os.path.join(navigator_config.CONF_PATH, 'Code')


def vscode_update_config(instance: 'base.BaseInstallableApp', prefix: str) -> None:  # pylint: disable=unused-argument
    """Update user config to use selected Python prefix interpreter."""
    try:
        _config_dir: str = os.path.join(navigator_config.CONF_PATH, 'Code', 'User')
        _config: str = os.path.join(_config_dir, 'settings.json')

        try:
            os.makedirs(_config_dir, exist_ok=True)
        except OSError as exception:
            logger.error(exception)
            return

        stream: typing.TextIO
        config_data: typing.Dict[str, typing.Any]
        if os.path.isfile(_config):
            try:
                with open(_config, 'rt', encoding='utf-8') as stream:
                    data = stream.read()
                vscode_create_config_backup(data)

                config_data = json.loads(data)
            except BaseException:  # pylint: disable=broad-except
                return
        else:
            config_data = {}

        pyexec: str = conda_api.get_pyexec(prefix)
        config_data.update({
            'python.terminal.activateEnvironment': True,
            'python.pythonPath': pyexec,
            'python.defaultInterpreterPath': pyexec,
            'python.condaPath': conda_api.get_pyscript(conda_api.CondaAPI().ROOT_PREFIX, 'conda'),
        })
        with open(_config, 'wt', encoding='utf-8') as stream:
            json.dump(config_data, stream, sort_keys=True, indent=4)

    except Exception as exception:  # pylint: disable=broad-except
        logger.error(exception)
        return


def vscode_create_config_backup(data: str) -> None:
    """
    Create a backup copy of the app configuration file `data`.

    Leave only the last 10 backups.
    """
    date: str = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    _config_dir: str = os.path.join(navigator_config.CONF_PATH, 'Code', 'User')
    _config_bck: str = os.path.join(_config_dir, f'bck.{date}.navigator.settings.json')

    # Make the backup
    stream: typing.TextIO
    with open(_config_bck, 'wt', encoding='utf-8') as stream:
        stream.write(data)

    # Only keep the latest 10 backups
    files: typing.List[str] = [
        os.path.join(_config_dir, item)
        for item in os.listdir(_config_dir)
        if item.startswith('bck.') and item.endswith('.navigator.settings.json')
    ]
    path: str
    for path in sorted(files, reverse=True)[10:]:
        try:
            os.remove(path)
        except OSError:
            pass


def vscode_install_extensions(instance: 'base.BaseInstallableApp') -> 'process.ProcessWorker':
    """Install app extensions."""
    if instance.executable is None:
        return instance._process_api.create_process_worker(['python', '--version'])  # pylint: disable=protected-access

    cmd: typing.Sequence[str] = [
        instance.executable,
        '--install-extension',
        'ms-python.anaconda-extension-pack',
        # 'ms-python-anaconda-extension',
        # 'ms-python.python',
    ]
    return instance._process_api.create_process_worker(cmd)  # pylint: disable=protected-access



# -*- coding: utf-8 -*-

"""A simple base to create singleton instances."""

from __future__ import annotations

__all__ = ['Singleton', 'SingleInstanceOf']

import abc
import typing

if typing.TYPE_CHECKING:
    import typing_extensions

T = typing.TypeVar('T')


__singletons__: 'typing_extensions.Final[typing.List[Singleton]]' = []


def reset_all() -> None:
    """Reset all singletons in application at once."""
    singleton: Singleton
    for singleton in __singletons__:
        singleton.reset()


class Singleton(typing.Generic[T], metaclass=abc.ABCMeta):
    """
    Container for a singleton instance of some type.

    Supports lazy initialization as soon as the instance might be required.
    """

    __slots__ = ('__instance',)

    def __init__(self) -> None:
        """Initialize new :class:`~Singleton` instance."""
        self.__instance: typing.Optional[T] = None
        __singletons__.append(self)

    @property
    def instance(self) -> T:  # noqa: D401
        """Singleton instance."""
        if self.__instance is None:
            self.__instance = self._prepare()
        return self.__instance

    def reset(self) -> None:
        """
        Reset singleton value.

        This will trigger repeated initialization of an instance on next :attr:`~Singleton.instance` call.
        """
        if self.__instance is None:
            return

        self._release()
        self.__instance = None

    @abc.abstractmethod
    def _prepare(self) -> T:
        """Initialize singleton instance."""

    def _release(self) -> None:
        """Destroy singleton instance when :meth:`~Singleton.reset` is called."""


class SingleInstanceOf(Singleton[T], typing.Generic[T]):
    """Shortcut for singletons constructed from external functions or instance constructors."""

    __slots__ = ('__constructor',)

    def __init__(self, constructor: typing.Callable[[], T]) -> None:
        """Initialize new :class:`~SingleInstanceOf` instance."""
        super().__init__()
        self.__constructor: 'typing_extensions.Final[typing.Callable[[], T]]' = constructor

    def _prepare(self) -> T:
        """Initialize singleton instance."""
        return self.__constructor()



# -*- coding: utf-8 -*-

"""Utilities to work with ANSI characters."""

__all__ = ['escape_ansi']

import re
import typing

if typing.TYPE_CHECKING:
    import typing_extensions

ESCAPE_ANSI_PATTERN: 'typing_extensions.Final[typing.Pattern[str]]' = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')


def escape_ansi(string: str) -> str:
    """
    Remove the ANSI escape sequences from a string.
    :param: String to cleared from ANSI escape sequences.
    :return: Cleared string.
    """
    return ESCAPE_ANSI_PATTERN.sub('', string)
